!2345678911234567892123456789312345678941234567895123456789612345678971234567898
!        0         0         0         0         0         0         0         0
!                   
! Module: attcalc_mod
!
! Authors:  Andy Read - XMM-SSC-MPE
!          
! Description: a
! 
! Comments: 
! 
! $Id: attcalc_mod.f90,v 1.3 2012/11/14 13:43:48 rsaxton Exp $
!- 
module attcalc_mod

  use cal
  use oal
  use dal
  use param  
  use errorHandling
  use emutils_mod
  
  implicit none 
  
  private 

  public attcalc

contains 
  
  !---------------------------------------------------------------------
  subroutine attcalc
  !---------------------------------------------------------------------

  type(SpacecraftAttitudeType)               :: att
  type(EulerAnglesType)                      :: euler
  type(DataSetT)                             :: infile, atthkfile
  type(TableT)                               :: inevents
  type(ColumnT)                              :: detxCol, detyCol, xCol, yCol, &
                                                timeECol

  character(len=*), parameter                :: taskname = "attcalc"
  character(len=256)                         :: history
  character(len=7)                           :: attsrc, instrument, datamode, &
                                                refpoint
 
  integer(kind=Int32), dimension(1)          :: dims
  integer(int32), dimension(:), pointer      :: p_x => null() , p_y => null()
  integer(kind=int32)                        :: calIval
  integer(int16), dimension(:), pointer      :: p_detx => null(), p_dety => null()
  integer(int16)                             :: cendetx, cendety
  integer                                    :: i, Ntimechange, &
                                                Nattbad, value, out_range, & 
                                                Nbadevents, ccdid, ccdnode, &
                                                Xmax, Ymax, Xmin, Ymin 

  real(kind=double), dimension(:), pointer   :: p_timeE => null()
  real(double), dimension(:,:), pointer      :: Cmatrix => null()
!  real(double), dimension(0:0)               :: nomxmm, nomymm, nomzmm
!  real(double), dimension(:), pointer        :: nomxmm1, nomymm1, nomzmm1
  integer(int16), dimension(0:0)             :: opticsx, opticsy
  real(double), dimension(:), pointer        :: opticsx_cc => null(), opticsy_cc => null(), &
                                                opticsz_cc => null()
!  integer(int16), dimension(:), pointer      :: nomoptx => null(), nomopty => null()
  real(double)                               :: &
                              attra, attdec, attPA, attraold, attdecold, &
                              attPAold, relx, rely, ranom, decnom, sdecnom, &
                              cdecnom, relx2, rely2, rahelp, dechelp, & 
                              ranomrad, attdecrad, rn, attPArad, diffPA, & 
                              xnew, ynew, rnew, bix, biy, bmtrx20, bmtrx21, &
                              srnew, crnew, attrahigh, attralow, attdechigh, &
                              attdeclow, attPAhigh, attPAlow, attrarad, &
                              attPAxmm, pixsize_y, bmtrx10, bmtrx11, &
                              ra_pnt, dec_pnt, pa_pnt, ix, iy, pixsize_x, &
                              bor1, bor2, bor3, sinr, cosr, rnb, relxb, &
                              relyb, calRval, pixelSize
  real                                       :: x, y, imagesize
                                          
  real, parameter                :: radian = 57.29577951E0, &
                                    skypixsize = 1.3888888888E-5, &!0.05arcsec, 
                                    pi = 3.1415926536E0

  logical                                     :: ranew, decnew, PAnew, &
                                                 attgood, timechange, mos, &
                                                 calBval, withatthk

  real(double)                :: tlmin_x, tlmax_x, tlmin_y, tlmax_y
!!!!! Needed for gfortran to pass the correct array descriptor to C++
#ifdef NAGf90Fortran
    real(kind=double), dimension(:,:), pointer :: Bmatrix => null()
#else
    real(kind=double), dimension(:,:), pointer :: Bmatrix => null()
   real(kind=double), dimension(3,3) :: tmpBmatrix
#endif

!******************************************************************************
!	START 
!
!------------------------------------------
!       GET INPUT FILE
!------------------------------------------
    infile = dataset(stringParameter("eventset"),MODIFY)
    if (.not.hasBlock(infile,"EVENTS")) then
      call error("badInfile","Input file incorrect - no EVENTS extension")
    endif
    inevents = table(infile,"EVENTS")
    if (.not.hasColumn(inevents,"TIME")) then
      call error("badInfile","Input file incorrect - no TIME column")
    endif
    if (.not.hasColumn(inevents,"DETX")) then
      call error("badInfile","Input file incorrect - no DETX column")
    endif
    if (.not.hasColumn(inevents,"DETY")) then
      call error("badInfile","Input file incorrect - no DETY column")
    endif
    timeECol = column(inevents,"TIME",READ)
    detxCol = column(inevents,"DETX",READ)
    detyCol = column(inevents,"DETY",READ)
    p_timeE => Real64Data(timeECol)
    p_detx => Int16Data(detxCol)
    p_dety => Int16Data(detyCol)
!------------------------------------------------
!	GET CAL/OAL and set state 
!------------------------------------------------
    mos = .false. 
    call CAL_setState(infile)
    call CAL_setState(accuracyLevel=1, randomize=.FALSE.)

    call CAL_getState(instrument=calIval)
    write (errstr,*) "Instrument = ",instId(calIval)
    call message(AppMsg,Verbose)
    instrument = instId(calIval)
    if (instrument == "EMOS1" .or. instrument == "EMOS2") mos = .true.
    call CAL_getState(ccdChipId=calIval)
    write (errstr,*) "CCD chip = ",calIval
    call message(AppMsg,Verbose)
    ccdid = calIval
    if (mos) then
      call CAL_getState(ccdNodeId=calIval)
      write (errstr,*) "CCD node = ",calIval
      call message(AppMsg,Verbose)
      ccdnode = calIval
    endif
    call CAL_getState(ccdModeId=calIval)
    write (errstr,*) "CCD mode = ",modeId(calIval)
    call message(AppMsg,Verbose)
    call CAL_getState(filterId=calIval)
    write (errstr,*) "Filter = ",filterId(calIval)
    call message(AppMsg,Verbose)
    call CAL_getState(accuracyLevel=calIval)
    write (errstr,*) "Accuracy level = ",calIval
    call message(AppMsg,Verbose)
    call CAL_getState(temperature=calRval)
    write (errstr,*) "Temperature = ",calRval
    call message(AppMsg,Verbose)
    call CAL_getState(temperatureCamera=calRval)
    write (errstr,*) "Camera Temperature = ",calRval
    call message(AppMsg,Verbose)
    call CAL_getState(randomize=calBval)
    write (errstr,*) "Randomization = ",calBval
    call message(AppMsg,Verbose)
!      call OAL_setState(inblock)
!      call CAL_setState(instId(stringAttribute(infile,"INSTRUME")))
    call OAL_setState(instId(instrument))
!      call OAL_setState(dataModeId(stringAttribute(infile,"DATAMODE")))
!------------------------------------------
!     GET INSTRUMENT 
!------------------------------------------
!    instrument = ChangeCase(stringAttribute(infile,"INSTRUME"),2)
    if (instrument /= "EPN".and. &
        instrument /= "EMOS1" .and. &
        instrument /= "EMOS2") then 
!        instrument /= "RGS1" .and. &
!        instrument /= "RGS2") then
      write (errstr,*) "Instrument ( ",instrument," ) not supported"
      call error("badInstrument")
    endif
    write(errstr,*) "Instrument: ",instrument
    call message(AppMsg,Verbose)
!-------------------------------------------
!     GET DATAMODE 
!-------------------------------------------
    datamode = ChangeCase(stringAttribute(infile,"DATAMODE"),2)
    if (datamode /= "IMAGING" .and. datamode /= "TIMING" .and. datamode /= "BURST") then
      write (errstr,*) "Mode ( ",datamode," ) not supported"
      call error("badDatamode")
    endif
    write(errstr,*) "Datamode: ",datamode
    call message(AppMsg,Verbose)
!-------------------------------------------
!     GET TCRPX and TCDLT keywords
!-------------------------------------------
    pixsize_x = real64Attribute(detxCol,"TCDLT")
    pixsize_y = real64Attribute(detyCol,"TCDLT")

!    cendetx = int16Attribute(detxCol,"TCRPX")
!    cendety = int16Attribute(detyCol,"TCRPX")


!
!   UL: The following code block computes the position of the optical
!       axes (from OPTICS_Y/OPTICS_Y in XMM_MISCDATA CCF) in the
!       CAMCOORD2 frame; the code block following (now commented out)
!       was errorneously believed to do the same
!
    call CAL_setState(ccdChipId= &
                          int(CAL_getMiscellaneousDataValue('OPTICS_CCD')))
    opticsx(0) = int(CAL_getMiscellaneousDataValue('OPTICS_X'))
    opticsy(0) = int(CAL_getMiscellaneousDataValue('OPTICS_Y'))

    call CAL_chipCoordToCamCoord1(opticsx, opticsy, &
                                  opticsx_cc, opticsy_cc, opticsz_cc);
    call CAL_camCoord1ToCamCoord2(opticsx_cc, opticsy_cc, opticsz_cc, &
                                  opticsx_cc, opticsy_cc, opticsz_cc);

    pixelSize = 3600.*skypixsize ! in arcsec
    ! opticsx/y_cc is in mm, so cendetxy is in
    !   mm * arcsec/pixel / mm/pixel = arcsec
    cendetx = int(opticsx_cc(0) * &
                  CAL_getMiscellaneousDataValue('PLATE_SCALE_X') / &
                  CAL_getMiscellaneousDataValue('MM_PER_PIXEL_X') / &
                    pixelSize + .5)
    cendety = int(opticsy_cc(0) * &
                  CAL_getMiscellaneousDataValue('PLATE_SCALE_Y') / &
                  CAL_getMiscellaneousDataValue('MM_PER_PIXEL_Y') / &
                    pixelSize + .5)

!    nomxmm(0) = 0.d0
!    nomymm(0) = 0.d0    ! center of CAMCOORD2 = nominal optical axis position
!    nomzmm(0) = 0.d0
!    call CAL_camCoord2ToCamCoord1(nomxmm, nomymm, nomzmm, &
!                                  nomxmm1, nomymm1, nomzmm1)
!    call CAL_camCoord1ToChipCoord(nomxmm1, nomymm1, nomzmm1, nomoptx, nomopty)
!    cendetx = int((nomoptx(0)-opticsx) * &
!                  CAL_getMiscellaneousDataValue('PLATE_SCALE_X') / &
!                    pixelSize + .5)
!    cendety = int((nomopty(0)-opticsy) * &
!                  CAL_getMiscellaneousDataValue('PLATE_SCALE_Y') / &
!                    pixelSize + .5)
    write(errstr,*) 'Position of optical axis in CAMCOORD2=', &
                cendetx, '/', cendety, ' * 0.05arcsec'
    call message(AppMsg,Noisy)
!    call CAL_releaseMemory(nomxmm1, nomymm1, nomzmm1)
!    call CAL_releaseMemory(nomoptx, nomopty)

!---------------------------------------------
!       ADD X/Y COLUMNS IF THEY DON'T EXIST (they should)
!---------------------------------------------
    if (.not.hasColumn(inevents,"X")) then 
      call message(AppMsg,Sparse," Column X does not exist - creating")
      dims(1)=1
      xCol = addColumn(inevents,"X",Integer32,"0.05 arcsec",dims)
    endif
    if (.not.hasColumn(inevents,"Y")) then 
      call message(AppMsg,Sparse," Column Y does not exist - creating")
      dims(1)=1
      yCol = addColumn(inevents,"Y",Integer32,"0.05 arcsec",dims)
    endif
    xCol = column(inevents,"X",MODIFY)
    yCol = column(inevents,"Y",MODIFY)
    p_x => Int32Data(xCol)
    p_y => Int32Data(yCol)
!------------------------------------------------- 
!       GET IMAGE SIZE (and convert to sky pixels) 
!-------------------------------------------------
    imagesize = realParameter("imagesize")   !NEW in degrees (def 0.36 = 21.6')
    if (imagesize /= 0.36) then 	     !different from default 
      write(errstr,*) "Note: imagesize chosen to be different from default value (0.36 deg)"
    endif
    imagesize = 72000.0 * imagesize  		!in 0.05arcs
    write(errstr,*) "Image size [in 0.05 arcs.] : ",imagesize
    call message(AppMsg,Verbose)
!---------------------------------------------------
!	GET SOURCE OF ATTITUDE INFO (ATTSRC)
!---------------------------------------------------
    attsrc = ChangeCase(stringParameter("attitudelabel"),2)
    if (attsrc /= "FIXED" .and. &
        attsrc /= "AHF" .and. &
        attsrc /= "OM") then
      write (errstr,*) &
        "Source of attitude information ( ",attsrc," ) incorrect - Should be FIXED, AHF or OM"
      call error("badAttsrc") 
    endif	
    write(errstr,*) "Source of attitude information: ",attsrc
    call message(AppMsg,Verbose)
!---------------------------------------------------
!	IF attsrc = FIXED, get attitude info 
!---------------------------------------------------
    if (attsrc == "FIXED") then
      attra = realParameter("fixedra")  
      attdec = realParameter("fixeddec")            
      attPA = realParameter("fixedposangle")            
    else 
!---------------------------------------------------
!	IF attsrc = AHF/OM, open OAL and set state 
!---------------------------------------------------
      if (attsrc == "AHF") then 
        call OAL_setState(attitudeFromAhf = .true.)       
      else if (attsrc == "OM") then
        call OAL_setState(attitudeFromAhf = .false.)       
      endif
    endif
!-------------------------------------------------
!	GET PNT values from atthkgen file 
!-------------------------------------------------
    withatthk = booleanParameter("withatthkset")
    if (withatthk) then
      if (setExists(stringParameter("atthkset"))) then ! atthkgen file exists
        atthkfile = dataset(stringParameter("atthkset"),READ)
        if (booleanParameter("withmedianpnt")) then 
          write(errstr,*) "Getting median PNT co-ords from atthkgen file: ",stringParameter("atthkset")
          call message(AppMsg,Verbose)
          if (attsrc == "AHF" .or. attsrc == "FIXED") then 
            ra_pnt = real64Attribute(atthkfile,"MAHFRA")
            dec_pnt = real64Attribute(atthkfile,"MAHFDEC")
            pa_pnt = real64Attribute(atthkfile,"MAHFPA")
          else if (attsrc == "OM") then
            ra_pnt = real64Attribute(atthkfile,"MOMRA")
            dec_pnt = real64Attribute(atthkfile,"MOMDEC")
            pa_pnt = real64Attribute(atthkfile,"MOMPA")
          endif
        else 
          write(errstr,*) "Getting mean PNT co-ords from atthkgen file: ",stringParameter("atthkset")
          call message(AppMsg,Verbose)
          if (attsrc == "AHF" .or. attsrc == "FIXED") then 
            ra_pnt = real64Attribute(atthkfile,"AAHFRA")
            dec_pnt = real64Attribute(atthkfile,"AAHFDEC")
            pa_pnt = real64Attribute(atthkfile,"AAHFPA")
          else if (attsrc == "OM") then
            ra_pnt = real64Attribute(atthkfile,"AOMRA")
            dec_pnt = real64Attribute(atthkfile,"AOMDEC")
            pa_pnt = real64Attribute(atthkfile,"AOMPA")
          endif
        endif
      endif
    endif
!-------------------------------------------------
!	GET CENTRAL REFERENCE POINT
!	refpoint now not redundant - many options (USER, NOMinal, OBJect, 
!       PoiNTing). [USER NOM OBJ PNT] NOM is default
!-------------------------------------------------
    refpoint = ChangeCase(stringParameter("refpointlabel"),2)
    if (refpoint /= "NOM" .and. refpoint /= "USER" .and. &
        refpoint /= "OBJ" .and. refpoint /= "PNT") then 
      write (errstr,*) &
        "Source of reference point information ( ",refpoint," ) incorrect - Should be NOM, USER, PNT or OBJ"
      call error("badRefpoint") 
    endif
    if (refpoint == "NOM") then 
      ranom = real64Attribute(infile,"RA_NOM")	     
      decnom = real64Attribute(infile,"DEC_NOM")     
    else if (refpoint == "OBJ") then
      ranom = real64Attribute(infile,"RA_OBJ")	     
      decnom = real64Attribute(infile,"DEC_OBJ")     
    else if (refpoint == "USER") then
      ranom = realParameter("nominalra")    
      decnom = realParameter("nominaldec")  
    else if (refpoint == "PNT") then
      if (ra_pnt > 399 .and. dec_pnt > 399 .and. pa_pnt > 399) then 
        call warning("badPntValue","PNT values in atthk file bad - using NOM values")
        ranom = real64Attribute(infile,"RA_NOM")	     
        decnom = real64Attribute(infile,"DEC_NOM")     
      else
        ranom = ra_pnt
        decnom = dec_pnt
      endif
      if (.not.setExists(stringParameter("atthkset")) .or. .not.withatthk) then ! no atthkgen file 
        if (withatthk) then 
          write(errstr,*) " No atthkgen file: ",trim(stringParameter("atthkset"))," - can't get PNT values - using NOM values"
          call warning("noAtthkgenFile")
        else
          call message(AppMsg,Verbose," Cannot access PNT values if withatthkset=N: using NOM values ")
        endif
        ranom = real64Attribute(infile,"RA_NOM")	     
        decnom = real64Attribute(infile,"DEC_NOM")     
        refpoint = "NOM"
      endif
    endif
    write(errstr,*) "Source of reference point information: ",refpoint
    call message(AppMsg,Verbose)
!---------------------------------------------------
!	CHECK ranom AND decnom ARE OK - GET sin AND cos
!---------------------------------------------------
    ranomrad = ranom/radian
    sdecnom = dsin(decnom/radian)
    cdecnom = dcos(decnom/radian)
    write(errstr,*) "Ref. point co-ords. ",ranom,decnom
    call message(AppMsg,Verbose)
!-------------------------------------------
!	Convert att info (fixed) to radians
!-------------------------------------------
    attrarad = attra/radian
    attdecrad = attdec/radian
    attPArad = attPA/radian
    !------------------------------------------------------
    ! 	INITIALIZE attraold, attdecold, attPAold, timechange etc.
    !------------------------------------------------------
    attraold   = 400.0
    attdecold  = 100.0
    attPAold = 400.0
    timechange = .TRUE.
    attgood = .TRUE.
    Ntimechange = 1
    Nbadevents = 0
    Nattbad = 0
    out_range = 0
    attrahigh = -0.1
    attralow = 6.3
    attdechigh = -3.2
    attdeclow = 3.2
    attPAhigh = -6.3
    attPAlow = 6.3
    Xmax = -99999999
    Ymax = -99999999
    Xmin = 99999999
    Ymin = 99999999
!-------------------------------------------
!	Set NULL values for X and Y
!-------------------------------------------
    call setNullValue(xCol,-99999999)
    call nullcheck(nullable(xCol))
    call setNullValue(yCol,-99999999)
    call nullcheck(nullable(yCol))
!------------------------------------------------------
!
!*****************  MAIN LOOP THROUGH EVENTS ***********
!
!------------------------------------------------------
    if (size(p_detx) < 1) then 
      call message(AppMsg,Sparse,"Events file contains no events")
    else
    do i = 0, size(p_detx)-1 		
  !----------------------------------------------!
  ! 	is time different? (apart from 1st)      !
  !----------------------------------------------!
      if (i > 0) then 
        if (abs(p_timeE(i) - p_timeE(i-1)) < 0.0001) then 
          timechange = .FALSE.
        else 
          timechange = .TRUE.
          Ntimechange = Ntimechange + 1
        endif
      endif
  !------------------------------------------------------!
  !  If no timechange go straight to event calculations  !
  !------------------------------------------------------!
    if (timechange) then
    !--------------------------------------------!
    !  If timechange, get ahf info (if needed)   !
    !--------------------------------------------!
    !--------------------------------!
    !  check if attitude info OK     !
    !--------------------------------!
      if (attsrc == "AHF" .or. attsrc == "OM") then 			
        if (OAL_getAttitude(p_timeE(i), att)) then 
          attrarad = att%ra   
          attdecrad = att%dec 
          attPArad = att%apos 
          attgood = .TRUE.
        else 			!attitude is bad
          attgood = .FALSE.
          Nattbad = Nattbad + 1
        endif
      endif
    !----------------------------------------------------------------------!
    !  has attitude info changed? (check which parts of info have changed) !
    !----------------------------------------------------------------------!
      if (attgood) then
        ranew = .false.
        decnew = .false.
        PAnew = .false.
        if (abs(attrarad - attraold) > 1.0d-7) then 
          attraold = attrarad
          ranew = .true.
          if (attrarad > attrahigh) attrahigh = attrarad 
          if (attrarad < attralow) attralow = attrarad 
        endif
        if (abs(attdecrad - attdecold) > 1.0d-7) then 
          attdecold = attdecrad
          decnew = .true.
          if (attdecrad > attdechigh) attdechigh = attdecrad 
          if (attdecrad < attdeclow) attdeclow = attdecrad 
        endif
        if (abs(attPArad - attPAold) > 1.0d-7) then 
          attPAold = attPArad
          PAnew = .true.
          if (attPArad > attPAhigh) attPAhigh = attPArad 
          if (attPArad < attPAlow) attPAlow = attPArad 
    !----------------------------------------------------------------------!
    !  Rotate PA by 90 (rosat to XMM) gives
    !      attPAxmm = attPArad - 1.570796327     !Xmm definition
    !  Astronomical PA is used (360 - PA) also, so ...
          attPAxmm = 4.712388980 - attPArad  
    !----------------------------------------------------------------------!    
        endif
    !--------------------------------------------------------------!
    !	transform from spherical to parallel co-ords (SUBR)        !
    !--------------------------------------------------------------!
        if (ranew .or. decnew) then 
          call spher2tang (attrarad, attdecrad, ranomrad, sdecnom, cdecnom, &
                           relx, rely)
        endif
    !--------------------------------------------------------------!
    !	do same 1deg away local north (help point) (SUBR) Ysc-axis !
    !--------------------------------------------------------------!
        if (ranew .or. decnew .or. PAnew) then
          call helppoint (attrarad, attdecrad, attPAxmm, & 
                          rahelp, dechelp)
          call spher2tang (rahelp, dechelp, ranomrad , sdecnom, cdecnom, & 
                           relx2, rely2)
    !-----------------------!     
    !	PA correction       ! 
    !-----------------------!
          diffPA = datan2(relx-relx2, rely2-rely)
          rn = sngl(4.71238898 - diffPA) !270 - diffpa; rn now in apa sense
          if (rn > 6.28319) rn = rn - 6.283185307
          if (rn < 0.0) rn = rn + 6.283185307
        endif
    !------------------------------------------!
    ! 	get boresight (SIAM) Bmatrix and angles         
    !------------------------------------------!
        call CAL_getBoresightMatrix(p_timeE(i), Bmatrix)
#ifdef NAGf90Fortran
    call OAL_toEulerAngles(Bmatrix, euler)
#else
!!!!! Needed for gfortran to pass the correct array descriptor to C++
!!!!! If this local variable is not created, the array descriptor we
!!!!! pass to C++ is not correct
    tmpBmatrix = Bmatrix
!!!!! Then, we have to transpose the matrix to preserve the order of the
!!!!! elements in C++. I know, it is ugly.
!!!!! This issue is also present in rgslib package.
    call OAL_toEulerAngles(transpose(Bmatrix), euler)

#endif

        !-----------------------------------------------------------!
        !  convert to bor3, bor2, bor1                              !
        !-----------------------------------------------------------!
        bor3 = euler%phi	!mv  pn4 ->pn7
        bor2 = euler%theta	!mv  pn4 ->pn1
        bor1 = euler%psi	!rtn pn4 ->pn5

! test errors in boresight (PN)
! 1 arcm = 2.908882E-04 radian
! bor3 = bor3 + (-0.25 * 2.908882E-04) 	!25/10/00 N253pn bore13
! bor2 = bor2 + (+0.075 * 2.908882E-04)	!        (also M31 good)
! bor1 = bor1 + (-1.75/radian)		!
! if (i == 0) write (*,*) "bor3, bor2, bor1"
! if (i == 0) write (*,*) "phi, theta, psi"
! if (i == 0) write (*,*) bor3,bor2,bor1
        call CAL_releaseMemory(Bmatrix)
        bor3 = dsin(bor3)		
        bor2 = dsin(bor2)
        bor1 = dsin(bor1)     !+0.0204238)  !test!!????!?!!!!!????
! test - no boresight
!        bor3 = 0.0 !dsin(pi)
!        bor2 = 0.0 !dsin(0.0)
!        bor1 = 0.0 !dsin(0.0)
        !---------------------------------------------!
        !   apply boresight correction (SUBR) to att  !
        !---------------------------------------------!
        sinr = dsin(dble(rn))
        cosr = dcos(dble(rn))
        call borxy (bor3, bor2, bor1, relx, rely, rn, relxb, relyb, rnb, &
                    sinr, cosr)
        rnew = rnb * radian			!rnew in degrees in apa sense
        srnew = sin(rnb)			!i.e. on sky anticlock (celN)
        crnew = cos(rnb)			!xnew, ynew in x/y frame
        xnew = relxb * radian/skypixsize	!all b/s corrected
        ynew = relyb * radian/skypixsize 	!
    !------------------------------------------!
    ! 	get  CC2->SC Cmatrix and angles        !??? move out of loop and rename
    !------------------------------------------!
        call CAL_getCamCoord2ToSacCoord(p_timeE(i), Cmatrix)
        bmtrx10 = Cmatrix(1,0)
        bmtrx11 = Cmatrix(1,1)
        bmtrx20 = Cmatrix(2,0)
        bmtrx21 = Cmatrix(2,1)
	!if (i == 0) write (*,*) Cmatrix(0,0),Cmatrix(0,1),Cmatrix(0,2)
	!if (i == 0) write (*,*) Cmatrix(1,0),Cmatrix(1,1),Cmatrix(1,2)
	!if (i == 0) write (*,*) Cmatrix(2,0),Cmatrix(2,1),Cmatrix(2,2)
        call CAL_releaseMemory(Cmatrix)
  !--------------------------------!
  !   If attgood is bad, come here !
  !--------------------------------!
        endif
  !-----------------------------------------!
  !   If no timechange, come straight here  !
  !-----------------------------------------!
      endif
  !--------------------------------------!
  ! If attitude is not good, flag event  !
  !--------------------------------------!
      if (.not.attgood) then 
        Nbadevents = Nbadevents + 1
        call setToNull(xCol,i)
        call setToNull(yCol,i)
      else
    !--------------------------------------!
    !	APPLY ATTITUDE INFO TO EVENT       ! 
    !--------------------------------------! 
        ix = (p_detx(i) - cendetx) * (pixsize_x/skypixsize) ! CORRECT
        iy = (p_dety(i) - cendety) * (pixsize_y/skypixsize) ! 
        bix = (bmtrx10 * ix) + (bmtrx11 * iy) ! = Ysc 
        biy = (bmtrx20 * ix) + (bmtrx21 * iy) ! = Zsc
        x = imagesize + 1 + xnew - (crnew*bix - srnew*biy) !minus for flips
        y = imagesize + 1 + ynew - (srnew*bix + crnew*biy) !from detback to sky
    !----------------------------------!
    !       PUT X Y IN EVENTS TABLE    !
    !----------------------------------!
        p_x(i) = nint(x)
        p_y(i) = nint(y)
        if (p_x(i) >= Xmax) Xmax = p_x(i)
        if (p_x(i) <= Xmin) Xmin = p_x(i)
        if (p_y(i) >= Ymax) Ymax = p_y(i)
        if (p_y(i) <= Ymin) Ymin = p_y(i)
      endif 
  !-------------------------------------------!
  !        CHECK TO SEE IF OUT OF IMAGE RANGE !
  !-------------------------------------------!
      if (x < 1 .or. &
        x > imagesize*2 .or. &
        y < 1 .or. & 
        y > imagesize*2) then
        out_range = out_range + 1
      endif
!---------------------------------------------------------
!
!*********************	END OF MAIN LOOP  ****************
    enddo     
!---------------------------------------------------------
    endif ! if more than zero events
!---------------------------------------------------------
!	WRITE OUT A FEW USEFUL THINGS
!---------------------------------------------------------
      write(errstr,*) "No of input events:        ", size(p_detx)
      call message(AppMsg,Verbose)
    if (attsrc /= "FIXED") then
      write(errstr,*) "No of event frames:        ", Ntimechange
      call message(AppMsg,Verbose)
      write(errstr,*) "No of bad attitude frames: ", Nattbad
      call message(AppMsg,Verbose)
      write(errstr,*) "No of bad attitude events: ", Nbadevents
      call message(AppMsg,Verbose)
      if (attrahigh >= attralow) then ! sensible ranges
        write(errstr,*) "Range in attitude RA:      ", attrahigh*radian, attralow*radian
        call message(AppMsg,Verbose)
        write(errstr,*) "Range in attitude DEC:     ", attdechigh*radian, attdeclow*radian
        call message(AppMsg,Verbose)
        write(errstr,*) "Range in attitude PA:      ", attPAhigh*radian, attPAlow*radian
        call message(AppMsg,Verbose)
      endif
    endif
    if (size(p_detx) > 0) then 
      write(errstr,*)   "1st event X,Y:             ", p_x(0), p_y(0)
      call message(AppMsg,Noisy)
      if (size(p_detx) > 3) then 
        write(errstr,*)   "2nd event X,Y:             ", p_x(1), p_y(1)
        call message(AppMsg,Noisy)
        write(errstr,*)   "3rd event X,Y:             ", p_x(2), p_y(2)
        call message(AppMsg,Noisy)
        write(errstr,*)   "4th event X,Y:             ", p_x(3), p_y(3)
        call message(AppMsg,Noisy)
      endif
    endif
!--------------------------------------------------------
! Output the number of events out of range if the calc

    if (out_range > 0 .and. .not. booleanParameter("calctlmax")) then
      write(errstr,*) out_range," events ( of ",size(p_detx)," ) out of range of image"
      call warning("outOfImageRange")
    endif

!--------------------------------------------
!  Calculate the TLMIN, TLMAX values for each axis. By default these are 
!  set to be the full image width. If requested then they should be calculated
!  on the basis of a minimum width of 0.36 degrees which is increased in units
!  of 3 arcmins.
!--------------------------------------------
    if (.not. booleanParameter("calctlmax")) then
       tlmin_x = 1
       tlmax_x = nint(2*imagesize)
       tlmin_y = 1
       tlmax_y = nint(2*imagesize)
    else
       write(*,*)' IN CALCTLMIN'
       tlmin_x = 1-3600*max(0,ceiling(-1.0*Xmin/3600.0))
       tlmax_x = 51840+3600*max(0,ceiling(float(Xmax-51840)/3600.0))
       tlmin_y = 1-3600*max(0,ceiling(-1.0*Ymin/3600.0))
       tlmax_y = 51840+3600*max(0,ceiling(float(Ymax-51840)/3600.0))
    endif

    write(errstr,*)'Set TLMIN,TLMAX to: ',int(tlmin_x),int(tlmax_x),int(tlmin_y),int(tlmax_y)
    call message(AppMsg,Verbose)

!--------------------------------------------
!  ADD WCS KEYWORDS
!--------------------------------------------
    call setAttribute(inevents,"EQUINOX",2.000000E+03,"","Equinox for sky coordinate x/y axes")
    call setAttribute(inevents,"RADECSYS","FK5","World coord. system for this file")
    call setAttribute(xCol,"TCTYP","RA---TAN","Coord. type: RA tangent plane projection")
      value = nint(imagesize + 1)
    call setAttribute(xCol,"TCRPX",value,"","axis reference pixel of projected image")
    call setAttribute(xCol,"TCRVL",ranom,"deg","coord. at X axis ref. pixel")
    call setAttribute(xCol,"TCDLT",-pixsize_x,"deg/pix","X increment at ref. pixel")
    call setAttribute(xCol,"TLMIN",tlmin_x,"","minimum legal QPOE projected image X axis value")
    call setAttribute(xCol,"TLMAX",tlmax_x,"","maximum legal QPOE projected image X axis value")
    call setAttribute(xCol,"TDMIN",Xmin,"","minimum projected image X axis data value")
    call setAttribute(xCol,"TDMAX",Xmax,"","maximum projected image X axis data value")
    call setAttribute(xCol,"TCUNI","deg","Physical units of X axis")
    call setAttribute(yCol,"TCTYP","DEC--TAN","Coord. type: DEC tangent plane projection")
      value = nint(imagesize + 1)
    call setAttribute(yCol,"TCRPX",value,"","axis reference pixel of projected image")
    call setAttribute(yCol,"TCRVL",decnom,"deg","coord. at Y axis ref. pixel")
    call setAttribute(yCol,"TCDLT",pixsize_y,"deg/pix","Y increment at ref. pixel")
    call setAttribute(yCol,"TLMIN",tlmin_y,"","minimum legal QPOE projected image Y axis value")
    call setAttribute(yCol,"TLMAX",tlmax_y,"","maximum legal QPOE projected image Y axis value")
    call setAttribute(yCol,"TDMIN",Ymin,"","minimum projected image Y axis data value")
    call setAttribute(yCol,"TDMAX",Ymax,"","maximum projected image Y axis data value")
    call setAttribute(yCol,"TCUNI","deg","Physical units of Y axis")
!----------------------------------------------------------
!	ADD attcalc ATTRIBUTES TO EXTENSION
!----------------------------------------------------------
    call setAttribute(inevents,"REFPOINT",refpoint,"source of reference point information")
    call setAttribute(inevents,"ATTSRC",attsrc,"source of attitude information")
    if (attsrc == "FIXED") then
      call setAttribute(inevents,"ATTRA",attra,"deg","fixed attitude RA")
      call setAttribute(inevents,"ATTDEC",attdec,"deg","fixed attitude Dec")
      call setAttribute(inevents,"ATTPA",attPA,"deg","fixed attitude position angle")
    endif  
!--------------------------------------------------------------
!	ADD ATTRIBUTES TO PRIMARY EVENTS HEADER also WCS copies
!--------------------------------------------------------------
    call setAttribute(infile,"EQUINOX",2.000000E+03,"","Equinox for sky coordinate x/y axes")
    call setAttribute(infile,"RADECSYS","FK5","World coord. system for this file")
    call setAttribute(infile,"REFXCTYP","RA---TAN","WCS Coord. type: RA tangent plane projection")
      value = nint(imagesize + 1)
    call setAttribute(infile,"REFXCRPX",value,"","WCS axis reference pixel of projected image")
    call setAttribute(infile,"REFXCRVL",ranom,"deg","WCS coord. at X axis ref. pixel")
    call setAttribute(infile,"REFXCDLT",-pixsize_x,"deg/pix","WCS X increment at ref. pixel")
    call setAttribute(infile,"REFXLMIN",1,"","WCS minimum legal QPOE projected image X axis value")
      value = nint(2*imagesize)
    call setAttribute(infile,"REFXLMAX",value,"","WCS maximum legal QPOE projected image X axis value")
    call setAttribute(infile,"REFXDMIN",Xmin,"","WCS minimum projected image X axis data value")
    call setAttribute(infile,"REFXDMAX",Xmax,"","WCS maximum projected image X axis data value")
    call setAttribute(infile,"REFXCUNI","deg","WCS Physical units of X axis")

    call setAttribute(infile,"REFYCTYP","DEC--TAN","WCS Coord. type: DEC tangent plane projection")
      value = nint(imagesize + 1)
    call setAttribute(infile,"REFYCRPX",value,"","WCS axis reference pixel of projected image")
    call setAttribute(infile,"REFYCRVL",decnom,"deg","WCS coord. at Y axis ref. pixel")
    call setAttribute(infile,"REFYCDLT",pixsize_y,"deg/pix","WCS Y increment at ref. pixel")
    call setAttribute(infile,"REFYLMIN",1,"","WCS minimum legal QPOE projected image Y axis value")
      value = nint(2*imagesize)
    call setAttribute(infile,"REFYLMAX",value,"","WCS maximum legal QPOE projected image Y axis value")
    call setAttribute(infile,"REFYDMIN",Ymin,"","WCS minimum projected image Y axis data value")
    call setAttribute(infile,"REFYDMAX",Ymax,"","WCS maximum projected image Y axis data value")
    call setAttribute(infile,"REFYCUNI","deg","WCS Physical units of Y axis")

    if (withatthk) then 
      if (.not.setExists(stringParameter("atthkset"))) then ! no atthkgen file 
        write(errstr,*) " No atthkgen file: ",trim(stringParameter("atthkset"))," - no PNT attributes propogated"
        call warning("noPntAttributes")
      else
        if (booleanParameter("withmedianpnt")) then 
          call setAttribute(infile,"AVRG_PNT","MEDIAN","Meaning of PNT values (mean or median)")
        else
          call setAttribute(infile,"AVRG_PNT","MEAN","Meaning of PNT values (mean or median)")
        endif
        call setAttribute(infile,"RA_PNT",ra_pnt,"deg","Actual (mean) pointing RA of the optical axis")
        call setAttribute(infile,"DEC_PNT",dec_pnt,"deg","Actual (mean) pointing Dec of the optical axis")
        call setAttribute(infile,"PA_PNT",pa_pnt,"deg","Actual (mean) measured position angle of the pointing axis")
      endif
    else 
      call message(AppMsg,Verbose," withatthkset=N: No PNT keywords propogated")
    endif

!----------------------------------------------
! Ensure that the event table header contains the same pointing keywords
! as the primary header
!----------------------------------------------
    call copyRefKeys(infile, inevents)

!----------------------------------------------
!	RELEASE EVENTS EXTENSION 
!----------------------------------------------
    call release(inevents)
    if (withatthk) then
      if (setExists(stringParameter("atthkset"))) then 
        call release(atthkfile) 
      endif
    endif
!-------------------------------------------------
!       ADD HISTORY (can put more information in)
!-------------------------------------------------
    history = "Source of reference point information: " // refpoint
    call addHistory(infile,history)
    history = "Source of attitude information: " // attsrc
    call addHistory(infile,history)
    if (attsrc == "FIXED") then
      write(history,'(a,f13.7,a)') "Attitude RA = ",attra
      call addHistory(infile,history)
      write(history,'(a,f13.7,a)') "Attitude DEC = ",attdec
      call addHistory(infile,history)
      write(history,'(a,f13.7,a)') "Attitude Pos.Angle = ",attPA
      call addHistory(infile,history)
    endif  
!--------------------------------------------------
!	RELEASE EVENTS FILE
!--------------------------------------------------
    call release(infile)
!--------------------------------------------------
!	*** END ***
    end subroutine attcalc
!--------------------------------------------------  

!
!	*** SUBROUTINES
!

  !---------------------------------------------------------------------
    subroutine spher2tang (attra, attdec, ranom, sdecnom, cdecnom, relx, &
                           rely)
    implicit none
    real(double)                     :: attra, attdec, sdecnom, &
                                        cdecnom, relx, rely, &
                                        ranom, diffra, dsinra, &
                                        dcosra, dsindec, dcosdec, &
                                        d1s, dh, d1c, dranew, d2s, d2c, &
                                        ddecnew

    diffra = attra - ranom
    dsinra = dsin(diffra)
    dcosra = dcos(diffra)
    dsindec = dsin(attdec)
    dcosdec = dcos(attdec)

!	residual alpha dranew
    d1s = dsinra * dcosdec
    dh = - dcosdec * dcosra
    d1c = sdecnom*dsindec - dh*cdecnom
    dranew = datan2(d1s,d1c)

! 	residual delta ddecnew
    d2s = dh*sdecnom + dsindec*cdecnom
    d2c = dsqrt(1.0d0 - d2s*d2s)
    ddecnew = datan2(d2s,d2c)

! 	convert to x,y in NS-system
    relx = - dranew
    rely =  - ddecnew
    end subroutine spher2tang
  !---------------------------------------------------------------------
    subroutine helppoint (attra, attdec, attPA, &
                           rahelp, dechelp)

    implicit none
    real, parameter        :: sin1deg = 1.74524064E-2, &
                              cos1deg = 9.99847695E-1
    real(double)           :: attra, attdec, attPA, &
                              rasin, decsin, PAsin, racos, deccos, &
                              PAcos, xhelp, yhelp, &
                              zhelp, rhelp, rahelp, &
                              dechelp

!	calc sines and cosines
      rasin = dsin(attra)
      racos = dcos(attra)
      decsin = dsin(attdec)
      deccos = dcos(attdec)
      PAsin = dsin(attPA)
      PAcos = dcos(attPA)

!	cartesian co-ords of help point
       xhelp = (sin1deg*(PAcos*racos*decsin - PAsin*rasin)) + &
                        (cos1deg*racos*deccos)
       yhelp = (sin1deg*(PAsin*racos + PAcos*rasin*decsin)) + &
                        (cos1deg*rasin*deccos)
       
       zhelp = - sin1deg*PAcos*deccos + cos1deg*decsin

!	length squared of help vector
      rhelp = xhelp*xhelp + yhelp*yhelp +zhelp*zhelp

      if (rhelp < 0.d0) then
!	error
        call warning("undefinedHelpVector","Help vector has undefined length - taking aspect solution as help point to contine")
        rahelp = attra
        dechelp = attdec
      else

!	spherical co-ords of help point
        rhelp = dsqrt(rhelp)
        rahelp = datan2 (yhelp, xhelp)
          if (rahelp < 0.d0) rahelp = rahelp + (2*3.1415926536)
        dechelp = dasin(zhelp/rhelp)
      endif

    end subroutine helppoint
  !---------------------------------------------------------------------
    subroutine borxy (bor3, bor2, bor1, relx, rely, rn, relxb, &
                      relyb, rnb, sinr, cosr)

    implicit none
    real(double)         :: bor3, bor2, bor1, relx, rely, rn, relxb, &
                            relyb, rnb, sinr, cosr, xbor, ybor

    xbor = 0.0
    ybor = 0.0
    xbor = (cosr * bor3) + (sinr * bor2) !in x/y frame
    ybor = (sinr * bor3) - (cosr * bor2) !in x/y frame

!       shift attitude vector by (xbor,ybor)

    relxb = relx + xbor
    relyb = - rely + ybor                !minus to put rely in x/y frame

!       correct "corrected Pos.Angle" also by fibor
! RDS CHANGE THIS SIGN OR NOT ??
!    rnb = rn + bor1   - ORIGINAL RETURN TO THIS
    rnb = rn - bor1  ! Corrected use of the psi angle - RDS (Oct 2015)

    end subroutine borxy
  !---------------------------------------------------------------------
    subroutine copyRefKeys(infile, inevents)

    type(DataSetT),intent(in)       :: infile
    type(TableT)                    :: inevents

! Copy the REF keywords - set in the main code
    call setAttribute(inevents,"REFXCTYP","RA---TAN","WCS Coord. type: RA tangent plane projection")

    call setAttribute(inevents,"REFXCRPX",int16Attribute(infile,"REFXCRPX"),"","WCS axis reference pixel of projected image")
    call setAttribute(inevents,"REFXCRVL",real64Attribute(infile,"REFXCRVL"),"deg","WCS coord. at X axis ref. pixel")
    call setAttribute(inevents,"REFXCDLT",real64Attribute(infile,"REFXCDLT"),"deg/pix","WCS X increment at ref. pixel")
    call setAttribute(inevents,"REFXLMIN",1,"","WCS minimum legal QPOE projected image X axis value")
    call setAttribute(inevents,"REFXLMAX",int16Attribute(infile,"REFXLMAX"),"",&
         "WCS maximum legal QPOE projected image X axis value")
    call setAttribute(inevents,"REFXDMIN",int16Attribute(infile,"REFXDMIN"),"","WCS minimum projected image X axis data value")
    call setAttribute(inevents,"REFXDMAX",int16Attribute(infile,"REFXDMAX"),"","WCS maximum projected image X axis data value")
    call setAttribute(inevents,"REFXCUNI","deg","WCS Physical units of X axis")

    call setAttribute(inevents,"REFYCTYP","DEC--TAN","WCS Coord. type: DEC tangent plane projection")
    call setAttribute(inevents,"REFYCRPX",int16Attribute(infile,"REFYCRPX"),"","WCS axis reference pixel of projected image")
    call setAttribute(inevents,"REFYCRVL",real64Attribute(infile,"REFYCRVL"),"deg","WCS coord. at Y axis ref. pixel")
    call setAttribute(inevents,"REFYCDLT",real64Attribute(infile,"REFYCDLT"),"deg/pix","WCS Y increment at ref. pixel")
    call setAttribute(inevents,"REFYLMIN",1,"","WCS minimum legal QPOE projected image Y axis value")
    call setAttribute(inevents,"REFYLMAX",int16Attribute(infile,"REFYLMAX"),"",&
         "WCS maximum legal QPOE projected image Y axis value")
    call setAttribute(inevents,"REFYDMIN",int16Attribute(infile,"REFYDMIN"),"","WCS minimum projected image Y axis data value")
    call setAttribute(inevents,"REFYDMAX",int16Attribute(infile,"REFYDMAX"),"","WCS maximum projected image Y axis data value")
    call setAttribute(inevents,"REFYCUNI","deg","WCS Physical units of Y axis")

! Also copy the PNT values if they have been set
    if (hasAttribute(infile,"RA_PNT") .and. hasAttribute(infile,"DEC_PNT") &
        .and. hasAttribute(infile,"PA_PNT")) then
       call setAttribute(inevents,"RA_PNT",real64Attribute(infile,"RA_PNT"),&
              "deg","Actual (mean) pointing RA of the optical axis")
       call setAttribute(inevents,"DEC_PNT",real64Attribute(infile,"DEC_PNT"),&
              "deg","Actual (mean) pointing Dec of the optical axis")
       call setAttribute(inevents,"PA_PNT",real64Attribute(infile,"PA_PNT"),&
              "deg","Mean  measured position angle of the pointing axis")
    endif

    end subroutine copyRefKeys

  !---------------------------------------------------------------------
    subroutine nullcheck (thisNullable)

    use dal
    type(NullableT),intent(in)  :: thisNullable
 
    if (.not.nullDefined(thisNullable)) then
      call warning("badXYNull","Null not defined correctly for X/Y columns")
    endif

    end subroutine nullcheck
  !---------------------------------------------------------------------
end module attcalc_mod

